#
# Copyright © 2020 United States Government as represented by the Administrator
# of the National Aeronautics and Space Administration. No copyright is claimed
# in the United States under Title 17, U.S. Code. All Other Rights Reserved.
#
# SPDX-License-Identifier: NASA-1.3
#
"""Bandpass filter curves."""
from astropy.table import QTable
from synphot import Empirical1D, SpectralElement

from importlib import resources

from . import constants
from . import data

__all__ = ('NUV_D', 'NUV_D_CBE', 'NUV_D_BASELINE', 'NUV_D_THRESHOLD')


def _get_bandpass(name):
    with resources.path(data, f'{name}_effective_area.ecsv') as p:
        t = QTable.read(p)
    return SpectralElement(Empirical1D,
                           points=t['wavelength'],
                           lookup_table=t['effective_area'] / constants.AREA)


NUV_D_CBE = _get_bandpass('nuv_d_cbe')
"""Effective area curve current best estimate."""

NUV_D_BASELINE = _get_bandpass('nuv_d_baseline')
"""Effective area curve for the baseline mission."""

NUV_D_THRESHOLD = _get_bandpass('nuv_d_threshold')
"""Effective area curve for the threshold mission."""

NUV_D = NUV_D_CBE
"""Default effective area curve (same as current best estimate)."""



#
# Copyright © 2021 United States Government as represented by the Administrator
# of the National Aeronautics and Space Administration. No copyright is claimed
# in the United States under Title 17, U.S. Code. All Other Rights Reserved.
#
# SPDX-License-Identifier: NASA-1.3
#
"""Sky backgrounds."""
from importlib import resources

from astropy.coordinates import GeocentricTrueEcliptic, get_sun, SkyCoord
from astropy.table import QTable
from astropy import units as u
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from synphot import Empirical1D, GaussianFlux1D, PowerLawFlux1D, SourceSpectrum
from synphot.models import ConstFlux1D
from synphot.units import PHOTLAM

from . import data

__all__ = ('get_zodiacal_light_scale', 'high_zodiacal_light', 'day_airglow',
           'get_airglow_scale', 'galactic1', 'galactic2',
           'get_galactic_scales')


def _get_zodi_angular_interp():
    # Zodiacal light angular dependence from Table 16 of Leinert et al. (2017)
    # https://doi.org/10.1051/aas:1998105.
    with resources.path(data, 'leinert_zodi.txt') as p:
        table = np.loadtxt(p)
    lat = table[0, 1:]
    lon = table[1:, 0]
    s10 = table[1:, 1:]

    # The table only extends up to a latitude of 75°. According to the paper,
    # "Towards the ecliptic pole, the brightness as given above is 60 ± 3 S10."
    lat = np.append(lat, 90)
    s10 = np.append(s10, np.tile(60.0, (len(lon), 1)), axis=1)

    # The table is in units of S10: the number of 10th magnitude stars per
    # square degree. Convert to magnitude per square arcsecond.
    sb = 10 - 2.5 * np.log10(s10 / 60**4)

    return RegularGridInterpolator([lon, lat], sb)


_zodi_angular_dependence = _get_zodi_angular_interp()


# Read zodiacal light spectrum
with resources.path(data, 'stis_zodi_high.ecsv') as p:
    table = QTable.read(p)
high_zodiacal_light = SourceSpectrum(
    Empirical1D,
    points=table['wavelength'],
    lookup_table=table['surface_brightness'] * u.arcsec**2)
""""High" zodiacal light spectrum, normalized to 1 square arcsecond.

The spectrum is from Table 6.4 of the STIS Instrument Manual.

References
----------
https://hst-docs.stsci.edu/stisihb/chapter-6-exposure-time-calculations/6-5-detector-and-sky-backgrounds

"""


def get_zodiacal_light_scale(coord, time):
    """Get the scale factor for zodiacal light compared to "high" conditions.

    Estimate the ratio between the zodiacal light at a specific sky position
    and time, and its "high" value, by interpolating Table 6.2 of the STIS
    Instrument Manual.

    Parameters
    ----------
    coord : astropy.coordinates.SkyCoord
        The coordinates of the object under observation. If the coordinates do
        not specify a distance, then the object is assumed to be a fixed star
        at infinite distance for the purpose of calculating its helioecliptic
        position.
    time : astropy.time.Time
        The time of the observation.

    Returns
    -------
    float
        The zodiacal light scale factor.

    References
    ----------
    https://hst-docs.stsci.edu/stisihb/chapter-6-exposure-time-calculations/6-5-detector-and-sky-backgrounds

    """
    obj = SkyCoord(coord).transform_to(GeocentricTrueEcliptic(equinox=time))
    sun = get_sun(time).transform_to(GeocentricTrueEcliptic(equinox=time))

    # Wrap angles and look up in table
    lat = np.abs(obj.lat.deg)
    lon = np.abs((obj.lon - sun.lon).wrap_at(180 * u.deg).deg)
    result = _zodi_angular_dependence(np.stack((lon, lat), axis=-1))

    # When interp2d encounters infinities, it returns nan. Fix that up here.
    result = np.where(np.isnan(result), -np.inf, result)

    # Fix up shape
    if obj.isscalar:
        result = result.item()

    result -= _zodi_angular_dependence([180, 0]).item()
    return u.mag(1).to_physical(result)


day_airglow = SourceSpectrum(
    GaussianFlux1D,
    mean=2471 * u.angstrom,
    fwhm=0.023 * u.angstrom,
    total_flux=1.5e-15 * u.erg * u.s**-1 * u.cm**-2)
"""Airglow spectrum in daytime, normalized to 1 square arcsecond."""


def get_airglow_scale(night):
    """Get the scale factor for the airglow spectrum.

    Estimate the ratio between the airglow and its daytime value.

    Parameters
    ----------
    night : bool
        Use the "low" value if True, or the "average" value if False.

    Returns
    -------
    scale : float
        The airglow scale factor.

    References
    ----------
    https://hst-docs.stsci.edu/stisihb/chapter-6-exposure-time-calculations/6-5-detector-and-sky-backgrounds

    """
    return np.where(night, 1e-2, 1)


galactic1 = SourceSpectrum(
    ConstFlux1D,
    amplitude=1 * PHOTLAM * u.steradian**-1 * u.arcsec**2)


galactic2 = SourceSpectrum(
    PowerLawFlux1D,
    x_0=1528*u.angstrom,
    alpha=-1,
    amplitude=1 * PHOTLAM * u.steradian**-1 * u.arcsec**2)


def get_galactic_scales(coord):
    """Get the Galactic diffuse emission, normalized to 1 square arcsecond.

    Estimate the Galactic diffuse emission based on the cosecant fits from
    Murthy (2014).

    Parameters
    ----------
    coord : astropy.coordinates.SkyCoord
        The coordinates of the object under observation.

    Returns
    -------
    synphot.SourceSpectrum
        The Galactic diffuse emission spectrum, normalized to 1 square
        arcsecond.

    References
    ----------
    https://doi.org/10.3847/1538-4357/aabcb9

    """
    b = SkyCoord(coord).galactic.b
    csc = 1 / np.sin(b)
    pos = (csc > 0)

    # Constants from Murthy (2014) Table 4.
    # Note that slopes for the Southern hemisphere have been negated to cancel
    # the minus sign in the Galactic latitude.
    fuv_a = np.where(pos, 93.4, -205.5)
    nuv_a = np.where(pos, 257.5, 66.7)
    fuv_b = np.where(pos, 133.2, -401.8)
    nuv_b = np.where(pos, 185.1, -356.3)

    fuv = fuv_a + fuv_b * csc
    nuv = nuv_a + nuv_b * csc

    # GALEX filter effective wavelengths in angstroms from
    # http://www.galex.caltech.edu/researcher/techdoc-ch1.html#3
    fuv_wave = 1528
    nuv_wave = 2271

    # return SourceSpectrum(
    #     PowerLawFlux1D,
    #     amplitude=fuv * surf_bright_unit,
    #     x_0=fuv_wave * u.angstrom,
    #     alpha=-np.log(nuv / fuv) / np.log(nuv_wave / fuv_wave))
    return fuv, (nuv - fuv) / (nuv_wave - fuv_wave)


del _get_zodi_angular_interp, table



#
# Copyright © 2021 United States Government as represented by the Administrator
# of the National Aeronautics and Space Administration. No copyright is claimed
# in the United States under Title 17, U.S. Code. All Other Rights Reserved.
#
# SPDX-License-Identifier: NASA-1.3
#
"""Dorado sensitivity calculator"""
try:
    from functools import cache
except ImportError:
    # FIXME: remove once we require Python 3.9 and higher.
    # functools.cache was added in Python 3.9, but is implemented using
    # lru_cache with a specific configuration. For our purposes of caching a
    # function that takes no arguments, functools.lru_cache can be used
    # interchangeably.
    from functools import lru_cache as cache

from astropy.stats import signal_to_noise_oir_ccd
from astropy import units as u
import numpy as np
from synphot import Observation
from synphot.units import PHOTLAM

from . import backgrounds
from . import bandpasses
from . import constants

__all__ = ('get_snr', 'get_limmag', 'get_exptime')


def _get_count_rate(source_spectrum, bandpass):
    flux = (source_spectrum * bandpass).integrate(flux_unit=PHOTLAM)
    return flux * (constants.AREA / u.ph)


def _get_background_count_rate(bandpass, coord, time, night):
    airglow_scale = backgrounds.get_airglow_scale(night)
    zodi_scale = backgrounds.get_zodiacal_light_scale(coord, time)
    galactic_scale1, galactic_scale2 = backgrounds.get_galactic_scales(coord)
    night = np.asarray(night)
    return np.square(constants.PLATE_SCALE.to_value(u.arcsec / u.pix)) * (
        _get_count_rate(backgrounds.high_zodiacal_light,
                        bandpass) * zodi_scale +
        _get_count_rate(backgrounds.day_airglow,
                        bandpass) * airglow_scale +
        _get_count_rate(backgrounds.galactic1,
                        bandpass) * galactic_scale1 +
        _get_count_rate(backgrounds.galactic2,
                        bandpass) * galactic_scale2
    )


@cache
def _get_dust_query():
    from dustmaps import planck
    planck.fetch(which='GNILC')
    return planck.PlanckGNILCQuery()


@cache
def _get_reddening_law():
    from dust_extinction.parameter_averages import F19
    from synphot import ReddeningLaw
    return ReddeningLaw(F19())


def _get_reddened_count_rate_scalar(source_spectrum, bandpass, ebv):
    reddening_law = _get_reddening_law()
    extinction_curve = reddening_law.extinction_curve(ebv, bandpass.waveset)
    return _get_count_rate(source_spectrum * extinction_curve, bandpass)


def _get_reddened_count_rate_vector(source_spectrum, bandpass, ebv):
    return u.Quantity([
        _get_reddened_count_rate_scalar(source_spectrum, bandpass, _)
        for _ in np.ravel(ebv)]).reshape(np.shape(ebv))


def _get_reddened_count_rate_slow(source_spectrum, bandpass, ebv):
    if np.isscalar(ebv):
        return _get_reddened_count_rate_scalar(source_spectrum, bandpass, ebv)
    else:
        return _get_reddened_count_rate_vector(source_spectrum, bandpass, ebv)


def _get_reddened_count_rate(source_spectrum, bandpass, ebv):
    steps = 100
    ebv_min = 0.0
    ebv_max = 50.0
    if np.size(ebv) < steps:
        return _get_reddened_count_rate_slow(source_spectrum, bandpass, ebv)
    else:
        from scipy.interpolate import interp1d
        x = np.linspace(ebv_min, ebv_max, steps)
        y = _get_reddened_count_rate_slow(source_spectrum, bandpass, x)
        unit = y.unit
        y = np.log(y.value)
        interp = interp1d(x, y, kind='cubic', assume_sorted=True,
                          bounds_error=False, fill_value=np.inf)
        return np.exp(interp(ebv)) * unit


def _get_source_count_rate(source_spectrum, bandpass, coord, redden):
    if redden:
        dust_query = _get_dust_query()
        ebv = dust_query(coord)
        return _get_reddened_count_rate(source_spectrum, bandpass, ebv)
    else:
        return _get_count_rate(source_spectrum, bandpass)


def get_snr(source_spectrum, *, exptime, coord, time, night, redden=False,
            bandpass=None):
    """Calculate the SNR of an observation of a point source with Dorado.

    Parameters
    ----------
    source_spectrum : synphot.SourceSpectrum
        The spectrum of the source.
    exptime : astropy.units.Quantity
        The exposure time
    coord : astropy.coordinates.SkyCoord
        The coordinates of the source, for calculating zodiacal light
    time : astropy.time.Time
        The time of the observation, for calculating zodiacal light
    night : bool
        Whether the observation occurs on the day or night side of the Earth,
        for estimating airglow
    redden : bool
        Whether to apply Milky Way extinction to the source spectrum
    bandpass : synphot.SpecralElement
        Bandpass. Default: Dorado current baseline estimate.

    Returns
    -------
    float
        The signal to noise ratio
    """
    if bandpass is None:
        bandpass = bandpasses.NUV_D
    return signal_to_noise_oir_ccd(
        exptime,
        constants.APERTURE_CORRECTION * _get_source_count_rate(
            source_spectrum, bandpass, coord, redden),
        _get_background_count_rate(bandpass, coord, time, night),
        constants.DARK_NOISE,
        constants.READ_NOISE,
        constants.NPIX
    ).to_value(u.dimensionless_unscaled)


def _amp_for_signal_to_noise_oir_ccd(
        snr, t, source_eps, sky_eps, dark_eps, rd, npix, gain=1.0):
    """Inverse of astropy.stats.signal_to_noise_oir_ccd."""
    signal = t * source_eps * gain
    # noise squared without signal shot noise term
    snr2 = np.square(snr)
    noise2 = t * (npix * (sky_eps * gain + dark_eps)) + npix * np.square(rd)
    return 0.5 * snr2 / signal * (1 + np.sqrt(1 + 4 * noise2 / snr2))


def get_limmag(source_spectrum, *, snr, exptime, coord, time, night,
               bandpass=None):
    """Get the limiting magnitude for a given SNR.

    Parameters
    ----------
    source_spectrum : synphot.SourceSpectrum
        The spectrum of the source.
    snr : float
        The desired SNR.
    exptime : astropy.units.Quantity
        The exposure time
    coord : astropy.coordinates.SkyCoord
        The coordinates of the source, for calculating zodiacal light
    time : astropy.time.Time
        The time of the observation, for calculating zodiacal light
    night : bool
        Whether the observation occurs on the day or night side of the Earth,
        for estimating airglow
    bandpass : synphot.SpecralElement
        Bandpass. Default: Dorado current baseline estimate.

    Returns
    -------
    astropy.units.Quantity
        The AB magnitude of the source
    """
    if bandpass is None:
        bandpass = bandpasses.NUV_D

    mag0 = Observation(source_spectrum, bandpass).effstim(
        u.ABmag, area=constants.AREA)

    result = _amp_for_signal_to_noise_oir_ccd(
        snr,
        exptime,
        constants.APERTURE_CORRECTION * _get_count_rate(
            source_spectrum, bandpass),
        _get_background_count_rate(bandpass, coord, time, night),
        constants.DARK_NOISE,
        constants.READ_NOISE,
        constants.NPIX
    ).to_value(u.dimensionless_unscaled)

    return -2.5 * np.log10(result) * u.mag + mag0


def _exptime_for_signal_to_noise_oir_ccd(
        snr, source_eps, sky_eps, dark_eps, rd, npix, gain=1.0):
    """Inverse of astropy.stats.signal_to_noise_oir_ccd."""
    c1 = source_eps * gain
    c2 = npix * (sky_eps * gain + dark_eps)
    c3 = npix * np.square(rd)
    x = 1 + c2 / c1
    snr2 = np.square(snr)
    return 0.5 * snr2 / c1 * (x + np.sqrt(np.square(x) + 4 * c3 / snr2))


def get_exptime(source_spectrum, *, snr, coord, time, night, redden=False,
                bandpass=None):
    """Calculate the SNR of an observation of a point source with Dorado.

    Parameters
    ----------
    source_spectrum : synphot.SourceSpectrum
        The spectrum of the source.
    snr : float
        The signal to noise ratio
    coord : astropy.coordinates.SkyCoord
        The coordinates of the source, for calculating zodiacal light
    time : astropy.time.Time
        The time of the observation, for calculating zodiacal light
    night : bool
        Whether the observation occurs on the day or night side of the Earth,
        for estimating airglow
    redden : bool
        Whether to apply Milky Way extinction to the source spectrum
    bandpass : synphot.SpecralElement
        Bandpass. Default: Dorado current baseline estimate.

    Returns
    -------
    astropy.units.Quantity
        The exposure time
    """
    if bandpass is None:
        bandpass = bandpasses.NUV_D
    return _exptime_for_signal_to_noise_oir_ccd(
        snr,
        constants.APERTURE_CORRECTION * _get_source_count_rate(
            source_spectrum, bandpass, coord, redden),
        _get_background_count_rate(bandpass, coord, time, night),
        constants.DARK_NOISE,
        constants.READ_NOISE,
        constants.NPIX
    )



#
# Copyright © 2020 United States Government as represented by the Administrator
# of the National Aeronautics and Space Administration. No copyright is claimed
# in the United States under Title 17, U.S. Code. All Other Rights Reserved.
#
# SPDX-License-Identifier: NASA-1.3
#
"""Scalar constants."""
from astropy import units as u
import numpy as np

AREA = 100.0 * u.cm**2
"""Fiducial collecting area.

Notes
-----
This is a fictitious area that is only used for converting effective area to
synphot dimensionless units. It is only equal to the actual collecting area of
the telescope to within an order of magnitude.
"""

PLATE_SCALE = 25.0 * u.arcsec * u.pix**-1
"""Plate scale.

Notes
-----
Assuing a fixed focal length and 2x2 binning.
"""

NPIX = np.pi * 0.89**2
"""Effective number of pixels in a circular aperture.

Notes
-----
26.8 um = 22.3" = 0.89 pix
"""

APERTURE_CORRECTION = 0.7
"""Aperture correction.

70% of the signal from a point source falls within the above circular aperture.
"""

DARK_NOISE = 4. * 30. * 0.124 * u.hour**-1
"""
Dark noise rate.

Notes
-----
From JPL spreadsheet; 4x for binning; 30x for MDF.
"""

READ_NOISE = 5.0
"""
Read noise.

Notes
-----
From CBE in proposal.
"""



#
# Copyright © 2020 United States Government as represented by the Administrator
# of the National Aeronautics and Space Administration. No copyright is claimed
# in the United States under Title 17, U.S. Code. All Other Rights Reserved.
#
# SPDX-License-Identifier: NASA-1.3
#
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy import units as u
from hypothesis import given, settings
from hypothesis.strategies import booleans, floats, just
from synphot import ConstFlux1D, SourceSpectrum
from pytest import approx

from .. import get_exptime, get_limmag, get_snr


@given(
    snr=floats(1, 1e9),
    exptime=floats(1, 1e9),
    ra=just(0),
    dec=just(90),
    time=just(Time('2020-01-01')),
    night=booleans())
@settings(deadline=None)
def test_round_trip_snr_limmag(snr, exptime, ra, dec, time, night):
    """Test round trip: get_snr(get_limmag(...))"""
    kwargs = dict(
        exptime=exptime * u.s,
        coord=SkyCoord(ra * u.deg, dec * u.deg),
        time=time, night=night)

    limmag = get_limmag(
        SourceSpectrum(ConstFlux1D, amplitude=0*u.ABmag), snr=snr, **kwargs)
    snr_2 = get_snr(SourceSpectrum(ConstFlux1D, amplitude=limmag), **kwargs)
    assert snr_2 == approx(snr)


@given(
    snr=floats(1, 1e9),
    mag=floats(15, 22),
    ra=just(0),
    dec=just(90),
    time=just(Time('2020-01-01')),
    night=booleans())
@settings(deadline=None)
def test_round_trip_snr_exptime(snr, mag, ra, dec, time, night):
    """Test round trip: get_snr(get_limmag(...))"""
    source = SourceSpectrum(ConstFlux1D, amplitude=mag * u.ABmag)
    kwargs = dict(
        coord=SkyCoord(ra * u.deg, dec * u.deg),
        night=night, time=time)

    exptime = get_exptime(source, snr=snr, **kwargs)
    snr_2 = get_snr(source, exptime=exptime, **kwargs)
    assert snr_2 == approx(snr)



#
# Copyright © 2020 United States Government as represented by the Administrator
# of the National Aeronautics and Space Administration. No copyright is claimed
# in the United States under Title 17, U.S. Code. All Other Rights Reserved.
#
# SPDX-License-Identifier: NASA-1.3
#
